from enum import Enum


class BalanceFormat(Enum):
    RAW = 1
    NANO = 10**30


def _truncate(number):
    # Improved to handle edge cases and ensure consistent output format
    return '{:.8f}'.format(number if number > 0 else 0.00)


def format_account_data(response, seed, index):
    return {
        "seed": seed,
        "index": index,
        "private": response["private"],
        "public": response["public"],
        "account": response["account"],
        "nano_prefix": response["account"][0:11],
        "nano_center": response["account"][11:59],
        "nano_suffix": response["account"][-6:]
    }


def format_balance_data(response, account):
    balance_format = BalanceFormat.NANO
    balance_raw = int(response["balance"])
    pending_raw = int(response["pending"])

    # Apply the multiplier based on the chosen format
    multiplier = balance_format.value
    balance = balance_raw / multiplier
    pending = pending_raw / multiplier
    total = (balance_raw + pending_raw) / multiplier

    # Format and return the results
    return {
        "account": account,
        "balance_raw": balance_raw,
        "balance": _truncate(balance),
        "pending": _truncate(pending),
        "total": _truncate(total)
    }



import traceback
import secrets
import json
import time
from nanomock.internal.utils import logger
from nanomock.modules.nl_nanolib import NanoLibTools, get_account_public_key
from nanorpc.client import NanoRpcTyped
from nanomock.modules.nl_rpc_utils import format_account_data, format_balance_data

# for block_creation, we store local frontier info, so that subsequant calls know about the most recent frontier without needing to publish the block to the ledger.
_FRONTIER_INFO = {}


class NanoRpc:
    def __init__(self, url, username=None, password=None, wrap_json=False):
        self.url = url
        self.nano_lib = NanoLibTools()
        self.nanorpc = NanoRpcTyped(
            url=url, username=username, password=password, wrap_json=wrap_json)

    def get_url(self):
        return self.url

    async def account_balance(self, account, include_only_confirmed):
        return await self.nanorpc.account_balance(account, include_only_confirmed=include_only_confirmed)

    async def account_info(self, account, representative=True, weight=None, receivable=True, pending=None, include_confirmed=True):
        return await self.nanorpc.account_info(account, representative=representative, weight=weight, receivable=receivable, pending=pending, include_confirmed=include_confirmed)

    async def accounts_balances(self, accounts, include_only_confirmed=None):
        return await self.nanorpc.accounts_balances(accounts, include_only_confirmed=include_only_confirmed)

    async def available_supply(self, ):
        return await self.nanorpc.available_supply()

    async def block_count(self, include_cemented=None):
        return await self.nanorpc.block_count(include_cemented=include_cemented)

    async def block_hash(self, block, json_block=True):
        return await self.nanorpc.block_hash(block, json_block=json_block)

    async def block_create(self, type_a, balance, key, representative, link, previous, work=None, version=None, difficulty=None, json_block=True):
        return await self.nanorpc.block_create(type_a, balance, key, representative, link, previous, work=work, version=version, difficulty=difficulty, json_block=json_block)

    async def block_info(self, block_hash, json_block=True):
        return await self.nanorpc.block_info(block_hash, json_block=json_block)

    async def deterministic_key(self, seed, index):
        return await self.nanorpc.deterministic_key(seed, index)

    async def key_create(self, ):
        return await self.nanorpc.key_create()

    async def key_expand(self, key):
        return await self.nanorpc.key_expand(key)

    async def process(self, block, force=None, subtype=None, json_block=True, async_=None):
        return await self.nanorpc.process(block, force=force, subtype=subtype, json_block=json_block, async_=async_)

    async def work_generate(self, block_hash, use_peers=None, difficulty=None, multiplier=None, account=None, version=None, block=None, json_block=True):
        return await self.nanorpc.work_generate(block_hash, use_peers=use_peers, difficulty=difficulty, multiplier=multiplier, account=account, version=version, block=block, json_block=json_block)

    async def account_create(self, wallet, index=None, work=None):
        return await self.nanorpc.account_create(wallet, index=index, work=work)

    async def account_list(self, wallet):
        return await self.nanorpc.account_list(wallet)

    async def receive(self, wallet, account, block):
        return await self.nanorpc.receive(wallet, account, block)

    async def wallet_create(self, seed=None):
        return await self.nanorpc.wallet_create(seed=seed)

    async def wallet_add(self, wallet, key, work=None):
        return await self.nanorpc.wallet_add(wallet, key, work=work)

    async def active_difficulty(self):
        return await self.nanorpc.active_difficulty()

    async def version(self):
        return await self.nanorpc.version()

    def generate_seed(self):
        return secrets.token_hex(32)

    async def check_balances(self, seed, start_index=0, end_index=50):
        result = []
        for index in range(start_index, end_index + 1):
            nano_account = await self.generate_account(seed, index)
            result.append(self.check_balance(nano_account["account"]))
        return result

    async def generate_account(self, seed, index):
        try:
            response = await self.deterministic_key(seed, index)
        except Exception as e:
            print(f"Error generating account from deterministic key: {e}")
            return None

        return format_account_data(response, seed, index)

    async def check_balance(self, account, include_only_confirmed=True):
        try:
            response = await self.account_balance(
                account, include_only_confirmed=include_only_confirmed)
        except Exception as e:
            print(f"Error fetching account balance: {e}")
            return None

        return format_balance_data(response, account)

    async def block_confirmed(self, json_block=True, block_hash=None):
        if json_block:
            block_hash = await self.block_hash(json_block)["hash"]
        if not block_hash:
            return False

        response = await self.block_info(block_hash)
        if response is None:
            return False
        if "error" in response:
            return False
        return True if response["confirmed"] == "true" else False

    async def create_open_block(self,
                                destination_account,
                                open_private_key,
                                amount_per_chunk_raw,
                                rep_account,
                                send_block_hash,
                                broadcast=True):
        block = await self.create_block("receive",
                                        source_private_key=open_private_key,
                                        destination_account=destination_account,
                                        representative=rep_account,
                                        amount_raw=amount_per_chunk_raw,
                                        link=send_block_hash,
                                        in_memory=not broadcast)

        return await self.get_block_result(block, broadcast)

    async def create_send_block(self,
                                source_seed,
                                source_index,
                                destination_account,
                                amount_per_chunk_raw,
                                broadcast=True):
        block = await self.create_block("send",
                                        source_seed=source_seed,
                                        source_index=source_index,
                                        destination_account=destination_account,
                                        amount_raw=amount_per_chunk_raw,
                                        in_memory=not broadcast)
        return await self.get_block_result(block,
                                           broadcast,
                                           source_seed=source_seed,
                                           source_index=source_index)

    async def create_change_block(self,
                                  source_seed,
                                  source_index,
                                  new_rep,
                                  broadcast=True):
        block = await self.create_block("change",
                                        source_seed=source_seed,
                                        source_index=source_index,
                                        link="0" * 64,
                                        representative=new_rep,
                                        in_memory=not broadcast)

        return await self.get_block_result(block,
                                           broadcast,
                                           source_seed=source_seed,
                                           source_index=source_index)

    async def create_change_block_pkey(self,
                                       source_private_key,
                                       new_rep,
                                       broadcast=True):
        block = await self.create_block("change",
                                        source_private_key=source_private_key,
                                        link="0" * 64,
                                        representative=new_rep,
                                        in_memory=not broadcast)
        return await self.get_block_result(block, broadcast)

    async def create_send_block_pkey(self,
                                     private_key,
                                     destination_account,
                                     amount_per_chunk_raw,
                                     broadcast=True):

        block = await self.create_block("send",
                                        source_private_key=private_key,
                                        destination_account=destination_account,
                                        amount_raw=amount_per_chunk_raw,
                                        in_memory=not broadcast)
        return await self.get_block_result(block, broadcast)

    async def create_epoch_block(self,
                                 epoch_link,
                                 genesis_private_key,
                                 genesis_account,
                                 broadcast=True):

        block = await self.create_block("epoch",
                                        source_private_key=genesis_private_key,
                                        destination_account=genesis_account,
                                        link=epoch_link,
                                        in_memory=not broadcast)

        return await self.get_block_result(block, broadcast)

    async def create_block(self,
                           sub_type,
                           link=None,
                           destination_account=None,
                           representative=None,
                           source_seed=None,
                           source_index=None,
                           source_private_key=None,
                           amount_raw=None,
                           in_memory=False,
                           add_in_memory=False,
                           read_in_memory=False,
                           use_rpc=True):
        try:
            if in_memory:
                add_in_memory = True
                read_in_memory = True
            if source_private_key is not None:
                source_account_data = self.nano_lib.nanolib_account_data(
                    private_key=source_private_key)
            elif source_seed is not None and source_index is not None:
                source_account_data = self.nano_lib.nanolib_account_data(
                    seed=source_seed, index=source_index)

            if read_in_memory:
                if source_account_data["account"] in _FRONTIER_INFO:
                    source_account_info = _FRONTIER_INFO[
                        source_account_data["account"]]
                else:
                    source_account_info = await self.account_info(
                        source_account_data["account"])
            else:
                source_account_info = await self.account_info(
                    source_account_data["account"])

            if representative is None:
                representative = source_account_info["representative"]
            if "balance" in source_account_info:
                balance = source_account_info["balance"]
            if "frontier" in source_account_info:
                previous = source_account_info["frontier"]

            if sub_type == "open" or sub_type == "receive":
                # destination_account = source_account_data["account"]
                if "error" in source_account_info:
                    sub_type = "open"
                    if use_rpc:
                        previous = "0" * 64
                    else:
                        previous = None
                    balance = amount_raw
                    # link = link
                else:
                    sub_type = "receive"
                    previous = source_account_info["frontier"]
                    balance = int(
                        source_account_info["balance"]) + int(amount_raw)
                    # link = link

            elif sub_type == "send":
                link = get_account_public_key(account_id=destination_account)
                balance = int(source_account_info["balance"]) - int(amount_raw)
                previous = source_account_info["frontier"]

            elif sub_type == "change":
                amount_raw = "0"
                destination_account = source_account_data["account"]
                # link = link

            elif sub_type == "epoch":
                if use_rpc:
                    pass
                else:
                    balance = int(source_account_info["balance"])

            if use_rpc:
                block = await self.block_create("state",
                                                balance,
                                                source_account_data["private"],
                                                representative,
                                                link,
                                                previous,
                                                json_block=True)
            else:
                active_difficulty = await self.active_difficulty()
                lib_block = self.nano_lib.create_state_block(
                    source_account_data["account"],
                    representative,
                    previous,
                    balance,
                    link,
                    source_account_data["private"],
                    difficulty=active_difficulty["network_minimum"])

                block = {
                    "hash": lib_block.block_hash,
                    "difficulty": lib_block.difficulty,
                    "block": json.loads(lib_block.json())
                }

            block["private"] = source_account_data["private"]
            block["subtype"] = sub_type
            block["amount_raw"] = amount_raw

            if "error" in block:
                block["success"] = False
                block["block"] = {}
                block["hash"] = None
            else:
                block["success"] = True
                block["error"] = None
                if add_in_memory:
                    _FRONTIER_INFO[source_account_data["account"]] = {
                        "frontier": block["hash"],
                        "balance": balance,
                        "representative": representative
                    }
            block["block"]["subtype"] = sub_type

        except Exception as e:
            traceback.print_exc()
            block = {
                "success": False,
                "block": {},
                "hash": None,
                "subtype": sub_type,
                "error": str(e)
            }
        return block

    async def get_block_result(self, block, broadcast, source_seed=None, source_index=None, exit_after_s=2):
        start_time = time.time()

        # Log an error immediately if the block was unsuccessful
        if not block.get("success", False):
            logger.warning(block.get("error", "Unknown error"))

        # Handle broadcast if needed
        if broadcast:
            publish = await self._try_publish_block(block, start_time, exit_after_s)
            broadcast = publish is not None and "hash" in publish
            if not broadcast:
                logger.error(
                    'Block not published: %s', block.get("hash", "Unknown hash"))

        # Construct the result dictionary
        result = {
            "success": block.get("success", False),
            "published": broadcast,
            "balance_raw": block["block"].get("balance", ""),
            "amount_raw": block.get("amount_raw", "0"),
            "hash": block.get("hash", ""),
            "block": block.get("block", {}),
            "subtype": block.get("subtype", ""),
            "account_data": {
                "account": block["block"].get("account", ""),
                "private": block.get("private", ""),
                "source_seed": source_seed,
                "source_index": source_index
            },
            "error": block.get("error", "")
        }

        # Log the entire result if the block was not successful
        if not result["success"]:
            logger.error(result)

        return result

    async def _try_publish_block(self, block, start_time, exit_after_s):
        while True:
            if time.time() - start_time > exit_after_s:
                return None
            publish = await self.process(block["block"], json_block=True)
            if publish is not None:
                return publish
            time.sleep(0.5)  # Sleep briefly before trying again



from typing import List, Dict, Optional, Tuple, Union
from math import floor
import json
import time
import inspect
import asyncio
import os
import shutil
from pathlib import Path
from asyncio import TimeoutError

import logging

from nanomock.internal.dependency_checker import DependencyChecker
from nanomock.modules.nl_parse_config import ConfigParser, ConfigReadWrite
from nanomock.internal.nl_initialise import InitialBlocks
from nanomock.docker import create_docker_interface
from nanomock.modules.nl_rpc import NanoRpc
from nanomock.internal.utils import log_on_success, shutil_rmtree, extract_packaged_services_to_disk, subprocess_run_capture_output
from nanomock.internal.utils import logger


class NanoLocalManager:

    def __init__(self, dir_path, project_name, config_file="nl_config.toml"):
        self.command_mapping = self._initialize_command_mapping()
        self.conf_p = ConfigParser(dir_path,
                                   config_file=config_file,
                                   logger=logger)
        self.conf_rw = ConfigReadWrite()
        self.dependency_checker = DependencyChecker()
        self.dependency_checker.check_dependencies()

        self.dir_path = dir_path
        self.nano_nodes_path = self.conf_p.nano_nodes_path
        self.compose_yml_path = self.conf_p.compose_out_path
        self.compose_env_path = os.path.join(self.nano_nodes_path,
                                             "dc_nano_local_env")
        self.project_name = project_name
        self.services_dir = self.conf_p.services_dir
        self.nodes_data_path = f"{self.nano_nodes_path}/{{node_name}}/NanoTest"
        self.config_node_path = f"{self.nano_nodes_path}/{{node_name}}/NanoTest/config-node.toml"
        self.config_rpc_path = f"{self.nano_nodes_path}/{{node_name}}/NanoTest/config-rpc.toml"
        self.config_log_path = f"{self.nano_nodes_path}/{{node_name}}/NanoTest/config-log.toml"

        os.makedirs(self.nano_nodes_path, exist_ok=True)

        self.docker_interface = create_docker_interface(
            self.compose_yml_path, self.project_name)

    def _initialize_command_mapping(self):
        # (command_method , validation_method)
        return {
            'create': (self.init_containers, None),
            'start': (self.start_containers, None),
            'start_nodes': (self.start_all_nodes, None),
            'status': (self.network_status, None),
            'restart': (self.restart_containers, None),
            'reset': (self.reset_nodes_data, None),
            'init': (self.init_nodes, None),
            'init_wallets': (self.init_wallets, None),
            'beta_create': (self.beta_create, None),
            'beta_init': (self.beta_init, None),
            'stop': (self.stop_containers, None),
            'stop_nodes': (self.stop_all_nodes, None),
            'remove': (self.remove_containers, None),
            'down': (self.remove_containers, None),
            'destroy': (lambda: self.destroy(remove_files=True), None),
            'rpc': (self.run_rpc, self._validator_rpc),
            'conf_edit': (self.conf_edit, self._validator_conf_edit)
        }

    def _get_default(self, config_name):
        # Mapping of config names to their respective default file names
        default_paths = {
            "config_node": "default_config-node.toml",
            "config_rpc": "default_config-rpc.toml",
            "config_log": "default_config-log.toml"
        }

        # Get the default path if the config name is valid
        default_file = default_paths.get(config_name)

        if default_file:
            default_path = os.path.join(self.services_dir, default_file)
            return self.conf_rw.read_toml(default_path, is_packaged=True)

        return {}

    def _generate_docker_compose_env_file(self):
        env_variables = self.conf_p.get_docker_compose_env_variables()
        self.conf_rw.write_list(f'{self.compose_env_path}', env_variables)

    def _generate_docker_compose_yml_file(self):
        self.conf_p.set_docker_compose()
        self.conf_p.write_docker_compose()

    def _set_config_log_file(self, node_name):
        config_log = self.conf_p.get_config_from_path(node_name, "config_log_path")
        if config_log is None:
            logger.debug( "No config-log.toml found. minimal version was created")
            config_log = self._get_default("config_log")
        config_log["log"]["default_level"] = self.conf_p.get_log_level(node_name)
        return config_log

    def _generate_config_log_file(self, node_name):
        config_log = self._set_config_log_file(node_name)
        self.conf_rw.write_toml(
            self.config_log_path.format(node_name=node_name), config_log)

    def _set_config_node_file(self, node_name):
        config_node = self.conf_p.get_config_from_path(node_name, "config_node_path")
        if config_node is None:
            logger.debug( "No config-node.toml found. minimal version was created")
            config_node = self._get_default("config_node")

        config_node["node"][
            "preconfigured_peers"] = self.conf_p.get_connected_peers(node_name)
        config_node["node"]["enable_voting"] = self.conf_p.is_voting_enabled(
            node_name)
        return config_node

    def _generate_config_node_file(self, node_name):
        config_node = self._set_config_node_file(node_name)
        self.conf_rw.write_toml(
            self.config_node_path.format(node_name=node_name), config_node)

    def _generate_config_rpc_file(self, node_name):
        config_rpc = self.conf_p.get_config_from_path(node_name, "config_rpc_path")
        if config_rpc is None:
            logger.debug( "No config-rpc.toml found. minimal version was created")
            config_rpc = self._get_default("config_rpc")

        self.conf_rw.write_toml(
            self.config_rpc_path.format(node_name=node_name), config_rpc)

    def _generate_nanomonitor_config_file(self, node_name):
        if self.conf_p.get_config_value("nanomonitor_enable"):
            self.conf_p.write_nanomonitor_config(node_name)

    def _create_node_folders(self, node_name):
        nano_node_path = os.path.join(self.nano_nodes_path, node_name)
        nano_test_path = os.path.join(nano_node_path, "NanoTest")

        os.makedirs(nano_node_path, exist_ok=True)
        os.makedirs(nano_test_path, exist_ok=True)

        if self.conf_p.get_config_value("nanomonitor_enable"):
            nano_monitor_path = os.path.join(nano_node_path, "nanoNodeMonitor")
            os.makedirs(nano_monitor_path, exist_ok=True)

    def _prepare_nodes(self, genesis_only=False):
        # prepare genesis
        nodes = self.conf_p.get_nodes_name()
        if genesis_only:
            nodes = [nodes[0]]
            logger.info("Only genesis node will be created")
        for node_name in self.conf_p.get_nodes_name():
            self._prepare_node_env(node_name)

    def _prepare_node_env(self, node_name):
        node_name = node_name.lower(
        )  # docker-compose requires lower case names
        self._create_node_folders(node_name)
        self._generate_config_node_file(node_name)
        self._generate_config_rpc_file(node_name)
        self._generate_config_log_file(node_name)
        self._generate_nanomonitor_config_file(node_name)

    def _online_containers(self, node_names: List[str]) -> List[str]:
        online_containers = []
        for container in node_names:
            cmd = f"docker ps |grep {container}$ | wc -l"
            res = subprocess_run_capture_output(cmd)
            online = int(res.stdout.strip())

            if online == 1:
                online_containers.append(container)

        return online_containers

    def _count_online_containers(self, node_names: List[str]) -> int:
        online_containers = self._online_containers(node_names)
        online_count = len(online_containers)
        return online_count

    async def _get_nodes_block_counts(self, nodes_name: List[str]) -> List[Dict[str, Union[str, int]]]:
        nodes_block_count = []

        nodes_rpc = ([
            NanoRpc(self.conf_p.get_node_rpc(node)) for node in nodes_name
        ] if nodes_name is not None else self._get_all_rpc())

        async def get_block_count_for_node(node_rpc):
            try:
                block_count = await asyncio.wait_for(node_rpc.block_count(), timeout=2.0)
                version_rpc_call = await asyncio.wait_for(node_rpc.version(), timeout=2.0)
                if not (block_count or version_rpc_call):
                    return None
                node_name = self.conf_p.get_node_name_from_rpc(node_rpc)
                count = int(block_count["count"])
                return {
                    "node_name": node_name,
                    "count": count,
                    "cemented": block_count["cemented"],
                    "version": version_rpc_call
                }
            except (asyncio.TimeoutError, Exception):
                return None

        results = await asyncio.gather(*(get_block_count_for_node(node_rpc) for node_rpc in nodes_rpc))
        nodes_block_count = [result for result in results if result is not None]
        return nodes_block_count



    def _generate_network_status_report(
            self, nodes: List[str],
            nodes_block_count: List[Dict[str, Union[str, int]]]) -> str:

        if not nodes_block_count:
            return ""

        max_count = max(int(bc["count"]) for bc in nodes_block_count)

        def get_node_data(
                node_name: str) -> Union[Dict[str, Union[str, int]], None]:
            for bc in nodes_block_count:
                if bc["node_name"] == node_name:
                    return bc
            return None

        def format_report_line(bc: Dict[str, Union[str, int]]) -> str:
            node_version = f'{bc["version"]["node_vendor"]} {bc["version"]["build_info"].split(" ")[0]}'
            synced_percentage = floor(
                int(bc["cemented"]) / max_count * 10000) / 100
            return '{:<16} {:<20} {:>6.2f}% synced | {}/{} blocks cemented'.format(
                bc["node_name"], node_version, synced_percentage,
                bc["cemented"], bc["count"])

        report = []
        for node_name in nodes:
            node_data = get_node_data(node_name)
            if node_data:
                report_line = format_report_line(node_data)
            else:
                report_line = f'{node_name} [down]'
            report.append(report_line)

        return '\n' + '\n'.join(report)


    def _get_all_rpc(self):
        return [NanoRpc(x) for x in self.conf_p.get_nodes_rpc()]

    async def _is_rpc_available(self,
                                container: str,
                                timeout: int = 3) -> Tuple[str, bool]:
        rpc_url = self.conf_p.get_node_rpc(container)
        try:
            nano_rpc = NanoRpc(rpc_url)
            block_count = await nano_rpc.block_count()
            if block_count:
                return container, True
        except Exception as e:
            logger.warning(
                f"RPC {rpc_url} not yet reachable for node {container}: {str(e)}"
            )

        return container, False

    async def _wait_for_rpc_availability(self,
                                         nodes_name: List[str] = None,
                                         wait: bool = True,
                                         timeout: int = 10,
                                         max_timeout_s: int = 45) -> None:
        start_time = time.time()
        nodes_name = nodes_name or self.conf_p.get_nodes_name()
        logger.warning(nodes_name)

        async def get_unavailable_nodes(nodes: List[str], timeout: int) -> List[str]:
            results = await asyncio.gather(*(self._is_rpc_available(node, timeout) for node in nodes))
            unavailable_nodes = [node for node, (node_name, available) in zip(
                nodes, results) if not available]
            return unavailable_nodes

        nodes_to_check = nodes_name.copy()

        while len(nodes_to_check) > 0:
            if not wait:
                break
            nodes_to_check = await get_unavailable_nodes(nodes_to_check, timeout)
            if time.time() - start_time > max_timeout_s:
                raise ValueError(
                    f"TIMEOUT: RPCs not reachable for nodes {nodes_to_check}")
            if len(nodes_to_check) > 0:
                await asyncio.sleep(1)  # Asynchronous sleep

        logger.info("Nodes %s reachable", nodes_name)

    def online_containers_status(self, online_containers: List[str],
                                 total_nodes) -> str:

        online_count = len(online_containers)
        if online_count == total_nodes:
            return f"All {online_count} containers online"
        else:
            return f"{online_count}/{total_nodes} containers online"

    async def conf_edit(self, payload):
        self.conf_p.modify_nanolocal_config(payload["path"], payload["value"])
        return True

    @log_on_success
    async def network_status(self, nodes_name: Optional[List[str]] = None) -> str:
        if nodes_name == []:
            return ""

        nodes_name = self.conf_p.get_nodes_name()
        online_containers = self._online_containers(nodes_name)
        status_msg = self.online_containers_status(online_containers,
                                                   len(nodes_name))
        # if "All" not in status_msg:
        #     return status_msg

        nodes_block_count = await self._get_nodes_block_counts(online_containers)
        return status_msg + self._generate_network_status_report(
            nodes_name, nodes_block_count)

    async def _create_docker_compose_file(self, genesis_only=False):
        extract_packaged_services_to_disk(self.nano_nodes_path)
        if genesis_only:
            genesis_name = self.conf_p.get_nodes_name()[0]
            self.conf_p.keep_nodes_by_name([genesis_name])
        self._prepare_nodes(genesis_only=genesis_only)
        self._generate_docker_compose_env_file()
        self._generate_docker_compose_yml_file()
        self.docker_interface.create_network(self.conf_p.get_network_name())
        self._initilaise_services_configs(self.conf_p)

        logger.success(
            f"Docker Compose file created at {self.compose_yml_path}")
        return None, "\n".join(self.conf_p.get_enabled_services())

    def _initilaise_services_configs(self, config: ConfigParser):
        if bool(config.get_config_value("nanocap_enable")):
            network_name = config.get_network_name()
            device_ip = self.docker_interface.get_network_gateway(network_name)
            config.set_nanocap_device_ip(device_ip)

    def _validator_rpc(self, nodes=None, payload=None):
        if not payload:
            raise ValueError(
                "The --payload argument is required for the 'rpc' command.")
        return nodes, payload

    def _validator_conf_edit(self, nodes=None, payload=None):
        if payload is None:
            raise ValueError(
                "payload must be provided '{\"path\" : ... , \"value\": ...}'")
        if "path" not in payload:
            raise ValueError("key \"path\" must be provided")
        if "value" not in payload:
            raise ValueError("key \"value\" must be provided")

        return nodes, payload

    @log_on_success
    async def run_rpc(self, payload=None, nodes=None):
        if nodes is None:
            nodes = self.conf_p.get_nodes_name()

        tasks = []
        for node in nodes:
            node_rpc = NanoRpc(self.conf_p.get_node_rpc(node))
            task = node_rpc.nanorpc.rpc.process_payloads([payload])
            tasks.append(task)

        responses = await asyncio.gather(*tasks)
        return None, json.dumps(responses, indent=2)


    @log_on_success
    async def init_wallets(self):
        init_blocks = InitialBlocks(self.conf_p,
                                    self.conf_p.get_nodes_rpc()[0])

        async def process_node(node_name):
            await self._wait_for_rpc_availability([node_name])

            if node_name == self.conf_p.get_genesis_node_name():
                await init_blocks.create_node_wallet(
                    self.conf_p.get_node_config(
                        self.conf_p.get_genesis_node_name())["rpc_url"],
                    self.conf_p.get_genesis_node_name(),
                    private_key=self.conf_p.config_dict["genesis_key"])
            else:
                await init_blocks.create_node_wallet(
                    self.conf_p.get_node_config(node_name)["rpc_url"],
                    node_name,
                    seed=self.conf_p.get_node_config(node_name)["seed"])

        node_names = self.conf_p.get_nodes_name()
        await asyncio.gather(*(process_node(node_name) for node_name in node_names))

        # return without logging for unit tests
        return init_blocks.logger.pop("InitialBlocks")

    async def _create_wallet(self, node_name, init_blocks):
        # Extract wallet creation logic into a separate coroutine
        if node_name == self.conf_p.get_genesis_node_name():
            await init_blocks.create_node_wallet(
                self.conf_p.get_node_config(self.conf_p.get_genesis_node_name())["rpc_url"],
                self.conf_p.get_genesis_node_name(),
                private_key=self.conf_p.config_dict["genesis_key"])
        else:
            await init_blocks.create_node_wallet(
                self.conf_p.get_node_config(node_name)["rpc_url"],
                node_name,
                seed=self.conf_p.get_node_config(node_name)["seed"])

    @log_on_success
    async def init_nodes(self):
        await self.init_wallets()
        init_blocks = InitialBlocks(self.conf_p,
                                    self.conf_p.get_nodes_rpc()[0])
        return await init_blocks.publish_initial_blocks()

    @log_on_success
    async def beta_create(self):
        await self.init_containers(genesis_only=True)

    @log_on_success
    async def beta_init(self):
        genesis_name = self.conf_p.get_nodes_name()[0]
        self.start_containers([genesis_name])
        init_blocks = InitialBlocks(self.conf_p,
                                    self.conf_p.get_nodes_rpc()[0])
        return init_blocks.publish_initial_blocks(move_weight=False)

    @log_on_success
    async def reset_nodes_data(self, nodes: Optional[List[str]] = None):
        await self.stop_containers(nodes)
        nodes_to_process = nodes or ['.']

        for node in nodes_to_process:
            node_path = Path(self.nano_nodes_path) / node if nodes else Path(
                self.nano_nodes_path)
            files_to_remove = ['data.ldb', 'wallets.ldb', 'rocksdb']

            for file_pattern in files_to_remove:
                for file_path in node_path.rglob(file_pattern):
                    if file_path.is_dir():
                        shutil.rmtree(file_path)
                    else:
                        file_path.unlink()

    @log_on_success
    async def init_containers(self, genesis_only=False):
        await self._create_docker_compose_file(genesis_only=genesis_only)
        await self.build_containers()

    @log_on_success
    async def restart_containers(self, nodes: Optional[List[str]] = None):
        return None, self.docker_interface.compose_restart(nodes)

    @log_on_success
    async def build_containers(self):
        return None, self.docker_interface.compose_build()

    @log_on_success
    async def start_containers(self, nodes: Optional[List[str]] = None):
        self.docker_interface.compose_start(nodes)
        await self._wait_for_rpc_availability(nodes)

    @log_on_success
    async def start_all_nodes(self):
        nodes = self.conf_p.get_nodes_name()
        self.docker_interface.compose_start(nodes)
        await self._wait_for_rpc_availability(nodes)

    @log_on_success
    async def stop_containers(self, nodes: Optional[List[str]] = None):
        # by default, stops all containers (also services like nanolooker, monitor or prom-exporter)
        self.docker_interface.compose_stop(nodes)

    @log_on_success
    async def stop_all_nodes(self):
        # stops all nodes but leaves services running
        nodes = self.conf_p.get_nodes_name()
        self.docker_interface.compose_stop(nodes)

    @log_on_success
    async def remove_containers(self):
        containers = self.conf_p.get_containers_name()
        if containers:
            self.docker_interface.compose_down()
        started_count = self.docker_interface.container_count(containers)
        removed_count = len(containers) - started_count
        return f"{removed_count} containers have been removed"

    @log_on_success
    async def destroy(self, remove_files=False):
        # Stop and remove containers
        await self.remove_containers()

        # Remove the created files and folders if remove_files is True
        if remove_files:
            return None, shutil_rmtree(self.nano_nodes_path)

    @log_on_success
    async def update(self):
        self.docker_interface.compose_pull()
        # Remove containers and networks
        await self.remove_containers()
        await self.build_containers()

    def _filter_args(self, func, **kwargs):
        sig = inspect.signature(func)
        filtered_args = {
            k: v
            for k, v in kwargs.items() if k in sig.parameters
        }
        return filtered_args

    async def execute_command(self, command, nodes=None, payload=None):
        if command not in self.command_mapping:
            raise ValueError(f"Invalid command: {command}")

        command_func, validator_func = self.command_mapping[command]

        if validator_func is not None:
            filtered_validator_args = self._filter_args(validator_func,
                                                        nodes=nodes,
                                                        payload=payload)
            validated_nodes, validated_payload = validator_func(
                **filtered_validator_args)
        else:
            validated_nodes, validated_payload = nodes, payload

        filtered_command_args = self._filter_args(command_func,
                                                  nodes=validated_nodes,
                                                  payload=validated_payload)
        await command_func(**filtered_command_args)


if __name__ == "__main__":
    pass
